{*******************************************************}
{                                                       }
{       MediaSource:TMediaSource komponenta 1.0         }
{                                                       }
{       Vyrobeno  (c) 2003, Z6 (zdeno6@seznam.cz)       }
{                                                       }
{*******************************************************}
unit MediaSource;

interface

uses
  Windows, Messages, SysUtils, Classes,ComCtrls,
  Forms,ZIPStream,Msk;

type
  TSourceType = (stFile,stZIP,stResource);

  TFindFileRec = record
    frFileName:string;
    frSourceType:TSourceType;
    frZIPName:string;
    frSourceName:string;
    frAllZIPs:Boolean;
    frAllItems:Boolean;
    SearchRec:TSearchRec;
    Case Boolean of
      true:
        (frFileHeader:PFileHeader);
      false:
        (frFile:PChar);
  end;

  PFindFileRec = ^TFindFileRec;

  TMediaPath = string;

  TSourceItem = class(TCollectionItem)
  private
    FName:string;
    FSourceType:TSourceType;
    FPath:TMediaPath;
    FZips:array of TZIP;
    FMem:TMemoryStream;
    procedure LoadBinProperty(Reader: TStream);
    procedure StoreBinProperty(Writer: TStream);
    procedure SetSourceType(Value:TSourceType);
    procedure SetPath(const Value:TMediaPath);
    procedure FreeZips;
    function GetPath:string;
    function GetZIP(Index:Integer):TZIP;
  protected
    function GetDisplayName: string; override;
    procedure DefineProperties(Filer: TFiler); override;
  public
    constructor Create(Collection: TCollection); override;
    destructor Destroy; override;
    function GetNamePath: string; override;
    function LoadFile(FileName:string;const ZipName:string):TStream;
    procedure InitZIPFiles;
    property ZIP[index:Integer]:TZIP read GetZIP;
  published
    property Name:string read FName write FName;
    property SourceType:TSourceType read FSourceType write SetSourceType;
    property Path:TMediaPath read GetPath write SetPath;
  end;

  TCustomMediaSource = class;
  TSourceList = class (TCollection)
  private
    FMediaSource:TCustomMediaSource;
  protected
    function GetAttrCount: Integer; override;
    function GetAttr(Index: Integer): string; override;
    function GetItemAttr(Index, ItemIndex: Integer): string; override;
  public
    function FindByName(const Name:string):TSourceItem;
  end;

  TCustomMediaSource = class(TComponent)
  private
    { Private declarations }
    FSourceList:TSourceList;
  protected
    { Protected declarations }
  public
    { Public declarations }
    constructor Create(AOwner:TComponent);override;
    destructor Destroy;override;
    property SourceList:TSourceList read FSourceList write FSourceList;
    function LoadFile(const SourceName:string;const FileName:string;const ZipName:string):TStream;
    function FindFirst(const SourceName:string;const FileName:string;const ZipName:string;var FindRec:TFindFileRec):Boolean;
    function FindNext(var FindRec:TFindFileRec):Boolean;
    procedure FindClose(var FindRec:TFindFileRec);
  end;

  TMediaSource = class (TCustomMediaSource)
  published
    property SourceList;
  end;

implementation

{ TCustomMediaSource }

constructor TCustomMediaSource.Create;
begin
  inherited;
  FSourceList:=TSourceList.Create(TSourceItem);
  FSourceList.FMediaSource:=Self;
end;

destructor TCustomMediaSource.Destroy;
begin
  FSourceList.Free;
  inherited;
end;

procedure TCustomMediaSource.FindClose(var FindRec: TFindFileRec);
begin
  with FindRec do
    {$WARNINGS OFF}
    if SearchRec.FindHandle <> 0 then begin
      SysUtils.FindClose(SearchRec);
      SearchRec.FindHandle:=0;
    end;
    {$WARNINGS ON}
end;

function TCustomMediaSource.FindFirst(const SourceName, FileName,
  ZipName: string; var FindRec: TFindFileRec): Boolean;
begin
  if (SourceList.Count = 0) or
     (SourceName<>'') and (SourceList.FindByName(SourceName)=nil) then begin
    Result:=false;
    exit;
  end;
  FillChar(FindRec,SizeOf(FindRec),0);
  with FindRec do begin
    if (SourceName = '') then begin
      frSourceName:=TSourceItem(SourceList.Items[0]).Name;
      frAllItems:=true;
    end else
      frSourceName:=SourceName;
    if ZIPName = '' then begin
      frZIPName:='';
      frAllZIPs:=true;
    end else if LastDelimiter('*?',ZIPName)>0 then
      frAllZIPs:=true;
      frZIPName:=ZIPName;
    frFileName:=FileName;
    frFile:=nil;
  end;
  Result:=FindNext(FindRec);
end;

function TCustomMediaSource.FindNext(var FindRec: TFindFileRec): Boolean;
var si:TSourceItem;
    zp:TZIP;
    i:Integer;
    findzip:Boolean;
  function Match(const name,mask:string):Boolean;
  begin
    // rozen o kontrolu cesty
    if SameText(ExtractFilePath(name),ExtractFilePath(mask)) then
      Result:=MatchesMask(name,mask)
    else
      Result:=false;
  end;
  procedure FindFileInZIP;
  var j:Integer;
  begin
    FindNext:=false;
    with FindRec do begin
      if frFile = nil then
        j:=-1     // prvn soubor
      else
        j:=zp.FindHeader(frFile);
      while j<Integer(zp.FileCount-1) do begin
        Inc(j);
        with PFileHeader(zp.Headers[j])^ do
          if Match(Filename,frFileName) then begin
            FindNext:=true;
            frZIPName:=zp.Name;
            frFile:=PChar(FileName);
            exit;
          end;
      end;
    end;
  end;
begin
  Result:=false;
  with FindRec do begin
    si:=SourceList.FindByName(frSourceName);
    while si <> nil do begin
      FindRec.frSourceName:=si.Name;
      FindRec.frSourceType:=si.SourceType;
      Case si.SourceType of
        stResource:
          begin
            zp:=si.ZIP[0];
            if (zp<>nil) and (zp.FileCount>0) then begin
              FindFileInZIP;
              if Result then exit;
            end;
            frFile:=nil;
          end;
        stZIP:
          begin
            i:=0;
            findzip:=true;  // zahaj hledn zipu podle frZIPName
            // nalezen vhodnho zipu
            while true do begin
              zp:=si.ZIP[i];
              if zp = nil then break;
              if (i = 0) and (frZIPName = '') then
                frZIPName:=zp.Name;
              if MatchesMask(zp.Name,frZIPName) and (zp.FileCount>0) then begin
                findzip:=false; // ZIP nalezen, te prohledme kad dal ZIP
                FindFileInZIP;
                if Result then exit;
                frFile:=nil;
              end;
              Inc(i);
              if not findzip and frAllZIPs and (si.ZIP[i]<>nil) then
                frZIPName:=si.ZIP[i].Name
            end;
          end;
        stFile:
          begin
            FindRec.frZIPName:='';
            if {$WARNINGS OFF} FindRec.SearchRec.FindHandle = 0 {$WARNINGS ON} then begin
              if SysUtils.FindFirst(si.Path+FindRec.frFileName,0,FindRec.SearchRec)=0 then begin
                FindRec.frFile:=PChar(FindRec.SearchRec.Name);
                Result:=true;
                exit;
              end;
            end else
            if SysUtils.FindNext(FindRec.SearchRec)=0 then begin
              FindRec.frFile:=PChar(FindRec.SearchRec.Name);
              Result:=true;
              exit;
            end;
          end;
      end;
      if frAllItems and (si.Index<SourceList.Count-1) then
        si:=TSourceItem(SourceList.Items[si.Index+1])
      else
        si:=nil;
      FindClose(FindRec);
    end;
  end;
end;

function TCustomMediaSource.LoadFile(const SourceName, FileName,
  ZipName: string): TStream;
var i:Integer;
    f:TSourceItem;
begin
  Result:=nil;
  if SourceName = '' then
    for i:=FSourceList.Count-1 downto 0 do begin
      Result:=TSourceItem(FSourceList.Items[i]).LoadFile(FileName,Zipname);
      if result<>nil then exit;
    end
  else begin
    f:=FSourceList.FindByName(SourceName);
    if f<>nil then
      Result:=f.LoadFile(FileName,ZipName);
  end;
end;

{ TSourceItem }

constructor TSourceItem.Create;
begin
  inherited;
  FPath:='*.*';
  FMem:=TMemoryStream.Create;
end;

procedure TSourceItem.DefineProperties(Filer: TFiler);
begin
  inherited;
  Filer.DefineBinaryProperty('Data',LoadBinProperty,StoreBinProperty,SourceType=stResource);
end;

destructor TSourceItem.Destroy;
begin
  FreeZips;
  FMem.Free;
  inherited;
end;

procedure TSourceItem.FreeZips;
var i:Integer;
begin
  for i:=0 to High(FZIPS) do with FZIPS[i] do begin
    if (ZipSource<>nil) and (ZipSource<>FMem) then
      ZipSource.Free;
    Free;
  end;
  FZIPS:=nil;
end;

function TSourceItem.GetDisplayName: string;
begin
  Result:=Name;
  if Result = '' then Result := GetNamePath;
end;

function TSourceItem.GetNamePath: string;
begin
  Result:=Name;
  if Result<>'' then exit;
  if Collection <> nil then
    Result := Format('%s[%d]',[Collection.GetNamePath, Index])
  else
    Result:=ClassName;
end;

procedure TSourceItem.InitZIPFiles;
var sr:TSearchRec;
    l:Cardinal;
    tz:TZIP;
    strm:TStream;
    nm:string;
    h:Integer;
begin
  FreeZips;
  Case SourceType of
    stZIP:
      begin
        with TSourceList(Collection).FMediaSource do
          if (csDesigning in ComponentState) then exit;
        if FindFirst(Path, 0, sr) = 0 then begin
          repeat
            nm:=ExtractFilePath(Path)+sr.Name;
            h:=FileOpen(nm,fmOpenRead or fmShareExclusive);
            FileRead(h,l,4);
            FileClose(h);
            if h<=0 then
              Continue;
            strm:=TStream(TFileStream.Create(nm,fmOpenRead or fmShareDenyNone));
            if (strm.Size>0) then begin
              if (strm.Size<98) then begin
                strm.Free;
                Continue;
              end else begin
                if l<>LOCHEADER_SIG then begin
                  strm.Free;
                  Continue;
                end;
              end;
            end;
            l:=Length(FZIPS);
            SetLength(FZIPS,l+1);
            tz:=TZIP.Create;
            tz.ZipSource:=strm;
            tz.Name:=sr.Name;
            FZIPS[l]:=tz;
          until FindNext(sr) <> 0;
          FindClose(sr);
        end;
      end;
    stResource:
      begin
          SetLength(FZIPS,1);
          tz:=TZIP.Create;
          tz.ZipSource:=FMem;
          FZIPS[0]:=tz;
      end;
  end;
end;

procedure TSourceItem.LoadBinProperty(Reader: TStream);
begin
  FMem.Size:=Reader.Size;
  if FMem.Size>0 then begin
    Reader.Read(FMem.Memory^,FMem.Size);
    InitZIPFiles;
  end;
end;

function TSourceItem.LoadFile(FileName:string; const ZipName: string): TStream;
var p:string;
    i:Integer;
begin
  Result:=nil;
  p:=ExtractFilePath(Path);
  case SourceType of
    stResource:
        if Length(FZIPS) > 0 then
          Result:=FZIPS[0].OpenFile(FileName);
    stFile:
        if FileExists(p+FileName) then begin
          Result:=TStream(TFileStream.Create(p+FileName,fmOpenRead));
        end;
    stZIP:
       for i:=0 to High(FZIPS) do
          if ((FZIPS[i].FindHeader(FileName)<>-1) and (ZipName='')) or
             (AnsiCompareText(FZIPS[i].Name,ZipName)=0)  then begin
            Result:=FZIPS[i].OpenFile(FileName);
          end;
  end;
end;

{procedure TSourceItem.SetReadOnly(Value: Boolean);
begin
  with TSourceList(Collection).FMediaSource do
  if (csDesigning in ComponentState) or
     (csLoading in ComponentState) then
    FReadOnly:=Value
  else
// tady by se hodila vyjmka
    exit;
end;
 }
procedure TSourceItem.SetSourceType(Value: TSourceType);
begin
  if Value = FSourceType then exit;
  if FSourceType <>stFile then
    FreeZIPs;
  with TSourceList(Collection).FMediaSource do
    if (csDesigning in ComponentState) then begin
      FSourceType:=Value;
      FMem.Size:=0;
      Case Value of
        stResource:
          Path:='';
        stZIP:
          Path:='*.zip';
      end;
    end else if (csLoading in ComponentState) then begin
      FSourceType:=Value;
//      exit;
    end else // za bhu lze pepnout pouze mezi stFile a stZIP
      if (SourceType=stResource) or (Value=stResource) then
        raise Exception.Create('SourceType error!')
      else
        FSourceType:=Value;
  InitZIPFiles;
end;

procedure TSourceItem.StoreBinProperty(Writer: TStream);
begin
//  if ZIP[0]<>nil then
    if FMem.Size > 0 then
      Writer.Write(FMem.Memory^,FMem.Size);
//    ZIP[0].SaveZIPToStream(Writer);
//    MessageBox(0,PChar('souboru '+IntToStr(ZIP[0].FileCount)),nil,0);
end;

function TSourceItem.GetPath: string;
begin
  Case FSourceType of
    stResource:
      Result:=ExtractFileName(Fpath);
    stFile:
      Result:=ExtractFilePath(Fpath);
    stZIP:
      Result:=FPath;
  end;
end;

function TSourceItem.GetZIP(Index: Integer): TZIP;
begin
  if Index>High(FZIPS) then
    Result:=nil
  else
    Result:=FZIPS[Index];
end;

procedure TSourceItem.SetPath(const Value: TMediaPath);
begin
  FPath:=Value;
  InitZIPFiles;
end;

{ TSourceList }

function TSourceList.FindByName(const Name: string): TSourceItem;
var i:Integer;
begin
  Result:=nil;
  for i:=0 to Count-1 do
    if CompareText(TSourceItem(Items[i]).FName,Name)=0 then begin
      Result:=TSourceItem(Items[i]);
      break;
    end;
end;

function TSourceList.GetAttr(Index: Integer): string;
begin
  Case Index of
    0:Result:='Jmno';
    1:Result:='Typ      ';
    2:Result:='Cesta';
    3:Result:='Velikost';
  else
    Result:='';
  end;
end;

function TSourceList.GetAttrCount: Integer;
begin
  Result:=4;
end;

function TSourceList.GetItemAttr(Index, ItemIndex: Integer): string;
begin
  Result:='';
  Case Index of
    0:Result:=Items[ItemIndex].DisplayName;
    1:Case TSourceItem(Items[ItemIndex]).SourceType of
        stFile:Result:='Soubor';
        stZIP:Result:='ZIP';
        stResource:Result:='Zdroj';
      end;
    2:Result:=TSourceItem(Items[ItemIndex]).Path;
    3:if TSourceItem(Items[ItemIndex]).SourceType = stResource then
        Result:=IntToStr(TSourceItem(Items[ItemIndex]).FMem.Size);
  end;
end;

end.
