{*******************************************************}
{                                                       }
{       TGA:natn obrzku ze souboru *.tga           }
{                                                       }
{       Vyrobeno  (c) 2003, Z6 (zdeno6@seznam.cz)       }
{                                                       }
{*******************************************************}
unit TGA;

interface

uses Windows,SysUtils,Graphics,Classes,Forms;

const
  TGA8Bit       = $01;
  TGAHighColor  = $02;
  TGA8bitCompressed = $09;
  TGACompressed = $0A;

type
  LogPal = record
    lpal : TLogPalette;
    dummy:Array[0..255] of TPaletteEntry;
  end;

  TTGAHeader = packed record
    IDLength:Byte;
    ColorMapType:Byte;
    ImageType:Byte;
    FirstColorMapEntry:Word;
    ColorMapLength:Word;
    ColorMapEntrySize:Byte;
    XOrigin,YOrigin:Word;
    Width:Word;
    Height:Word;
    BitPerPixel:Byte;
    Descriptor:Byte;
  end;

  TTGAImage = class (TGraphic)
  private
    FBitmap:TBitmap;
    FAlfa:TBitmap;
    FGetAlfa:Boolean; // true = vrac alfa vrstvu a ne RGB
    FCompressed:Boolean; // pvodn zznam byl komprimovan
    function GetPixelFormat:TPixelFormat;
  protected
    procedure AssignTo(Dest: TPersistent); override;
    function GetEmpty: Boolean; override;
    function GetHeight: Integer; override;
    function GetWidth: Integer; override;
    procedure SetHeight(Value: Integer); override;
    procedure SetWidth(Value: Integer); override;
    procedure Draw(ACanvas: TCanvas; const Rect: TRect); override;
  public
    property PixelFormat:TPixelFormat read GetPixelFormat;
    property Alfa:Boolean read FGetAlfa write FGetAlfa;
    property Compressed:Boolean read FCompressed;
    procedure Assign(Source: TPersistent); override;
    constructor Create; override;
    destructor Destroy;override;
    procedure LoadFromStream(Stream: TStream); override;
    procedure SaveToStream(Stream: TStream); override;
    procedure LoadFromClipboardFormat(AFormat: Word; AData: THandle;
      APalette: HPALETTE); override;
    procedure SaveToClipboardFormat(var AFormat: Word; var AData: THandle;
      var APalette: HPALETTE); override;
  end;

implementation

{ TTGAImage }

procedure TTGAImage.Assign(Source: TPersistent);
begin
  if Source is TTGAImage then begin
      FBitmap.Assign(TTGAImage(Source).FBitmap);
      FAlfa.Free;
      FAlfa:=nil;
      if TTGAImage(Source).FAlfa<>nil then begin
        FAlfa:=TBitmap.Create;
        FAlfa.Assign(TTGAImage(Source).FAlfa);
        FGetAlfa:=TTGAImage(Source).FGetAlfa;
      end;
  end else if Source is TBitmap then
    FBitmap.Assign(Source)
  else
    inherited Assign(Source);
end;

procedure TTGAImage.AssignTo(Dest: TPersistent);
begin
  if Dest is TBitmap then begin
    if FGetAlfa and (FAlfa<>nil) then
      Dest.Assign(FAlfa)
    else
      Dest.Assign(FBitmap)
  end else
    inherited AssignTo(Dest);
end;

constructor TTGAImage.Create;
begin
  inherited;
  FBitmap:=TBitmap.Create;
//  FBitmap.LoadFromFile(ExtractFilePath(Application.ExeName)+'\Close.bmp');
  FAlfa:=nil;
end;

destructor TTGAImage.Destroy;
begin
  FBitmap.Free;
  if FAlfa<>nil then
    FAlfa.Free;
  inherited;
end;

procedure TTGAImage.Draw(ACanvas: TCanvas; const Rect: TRect);
begin
  if FGetAlfa and (FAlfa <> nil) then
    ACanvas.StretchDraw(Rect,FAlfa)
  else
    ACanvas.StretchDraw(Rect,FBitmap)
end;

function TTGAImage.GetEmpty: Boolean;
begin
  Result:=FBitmap.Empty;
end;

function TTGAImage.GetHeight: Integer;
begin
  if FBitmap<>nil then
    Result:=FBitmap.Height
  else
    Result:=0;
end;

function TTGAImage.GetWidth: Integer;
begin
  if FBitmap<>nil then
    Result:=FBitmap.Width
  else
    Result:=0;
end;

procedure TTGAImage.LoadFromClipboardFormat(AFormat: Word; AData: THandle;
  APalette: HPALETTE);
begin
end;

procedure TTGAImage.SaveToClipboardFormat(var AFormat: Word;
  var AData: THandle; var APalette: HPALETTE);
begin
end;

procedure TTGAImage.SaveToStream(Stream: TStream);
begin
end;

procedure TTGAImage.SetHeight(Value: Integer);
begin
end;

procedure TTGAImage.SetWidth(Value: Integer);
begin
end;

procedure TTGAImage.LoadFromStream(Stream: TStream);
type pRGB = packed record
       B,G,R:byte;
     end;
     pRGBA = packed record
       B,G,R,A:byte;
     end;
var h:TTGAHeader;
    i,j:Integer;
    buffer:Pointer;
    size:Integer;
    tempsize:Integer;
    filepal:array [0..255] of pRGB;
    SysPal:LogPal;
    pixel:^pRGBA;
    dest:PByte;
    prefix:Byte;
    pixbuf:Cardinal;
    bytes:Byte;
begin
  if FAlfa<>nil then begin
    FAlfa.Free;
    FAlfa:=nil;
  end;
  // naten hlaviky
  Stream.Read(h,SizeOf(h));
  if h.BitPerPixel = 8 then
    FBitmap.PixelFormat:=pf8bit
  else if h.BitPerPixel = 16 then
    FBitmap.PixelFormat:=pf15bit
  else if h.BitPerPixel = 24 then
    FBitmap.PixelFormat:=pf24bit
  else if h.BitPerPixel = 32 then begin
    FBitmap.PixelFormat:=pf32bit;
    // vytvoen alfa masky
    FAlfa:=TBitmap.Create;
    FAlfa.PixelFormat:=pf8bit;
    FAlfa.Width:=h.Width;
    FAlfa.Height:=h.Height;
  end;
  FBitmap.Width:=h.Width;
  FBitmap.Height:=h.Height;
  if h.ColorMapType <>0 then begin
    // naten palety
    Stream.Read(filepal,3*h.ColorMapLength);
    SysPal.lPal.palVersion:=$300;
    SysPal.lPal.palNumEntries:=h.ColorMapLength;
    for i:=0 to h.ColorMapLength-1 do
      with SysPal.lpal.palPalEntry[i],filepal[i] do begin
        peRed:=R ;
        peGreen:=G ;
        peBlue:=B;
        peFlags:=0;
      end;
    FBitmap.Palette:= CreatePalette(Syspal.lpal);
  end;
  bytes:=(h.BitPerPixel+7) div 8;
  size:=bytes * h.Width;
  buffer:=AllocMem(size);
  FCompressed:=(h.ImageType and 8) <> 0;
  try
    for i:=0 to h.Height-1 do begin
      if not FCompressed then
        Stream.Read(buffer^,size) // bez komprimace pmo nateme cel dek
      else begin  // nejprve musme dekomprimovat
        tempsize:=size;
        dest:=buffer;
        while tempsize>0 do begin // dokud nezskme cel dek pak
          Stream.Read(prefix,SizeOf(prefix));  // nati prefix
          Stream.Read(pixbuf,bytes);           // nati nsledujc pixel (velikost zvis na potu bit na jeden pixel
          for j:=(prefix and $7f) downto 0 do begin  // pak zapi (prefix and $7F) + 1 pixel
            Move(pixbuf,dest^,bytes);
            Inc(dest,bytes);
            Dec(tempsize,bytes);
            if tempsize<=0 then break;
            if (j>0) and ((prefix and $80) = 0) then  // pokud prefix <128, pak se nataj dal pixely, jinak se pouije 1.naten
              Stream.Read(pixbuf,bytes);
          end;
        end;
      end;
      if (h.Descriptor and $20)<>0 then  // zapsno od nejvrchnjho dku
        Move(buffer^,FBitmap.ScanLine[i]^,size)
      else                // zapsno od nejspodnjho dku
        Move(buffer^,FBitmap.ScanLine[Height-1-i]^,size);
    end;
  finally
    FreeMem(buffer);
  end;
  // pekoprovn informac v alfa kanle do FAlfa
  if FAlfa<>nil then begin
    // nastaven ed palety
    SysPal.lPal.palVersion:=$300;
    SysPal.lPal.palNumEntries:=256;
    for i:=0 to 255 do
      with SysPal.lpal.palPalEntry[i] do begin
        peRed:=not i ;
        peGreen:=not i ;
        peBlue:=not i;
        peFlags:=0;
      end;
    FAlfa.Palette:= CreatePalette(Syspal.lpal);
    for i:=h.Height-1 downto 0 do begin
      pixel:=FBitmap.ScanLine[i];
      dest:=FAlfa.ScanLine[i];
      for j:=0 to h.Width-1 do begin
        dest^:=pixel^.A;
        Inc(pixel);
        Inc(dest);
      end;
    end;
  end;
{  else
    FBitmap.FreeImage;
    FBitmap.Width:=0;
    FBitmap.Height:=0;
  end;}
end;

function TTGAImage.GetPixelFormat: TPixelFormat;
begin
  if FBitmap<>nil then
    Result:=FBitmap.PixelFormat
  else
    Result:=pfCustom;
end;

end.
