{*******************************************************}
{                                                       }
{       ZIPStream:Prce se zipy                         }
{                                                       }
{       Vyrobeno  (c) 2003, Z6 (zdeno6@seznam.cz)       }
{                                                       }
{*******************************************************}
unit ZIPStream;

interface

uses Windows,Classes,SysUtils,CalcCrc;

const
   // signatura jednotlivch blok
   LOCHEADER_SIG : Cardinal = $04034b50;
   HEADER_SIG    : Cardinal = $02014b50;
   DIGITAL_SIG   : Cardinal = $05054b50;
   ENDOFDIR_SIG  : Cardinal = $06054b50;

   // velikost hlaviek jednotlivch blok
   LOCHEADERSIZE = 30;
   HEADERSIZE = 46;
   DIRENDSIZE = 22;

   zlib_Version = '1.0.4';
type
  TCompressionLevel = (clNone, clFastest, clDefault, clMax);

  // struktura lokln hlaviky
  TLocalHeader = packed record
    Version                     : Word;
    Flag                        : Word;
    CompressionMethod           : Word;
    DateTime                    : Cardinal; { Time: Word; Date: Word; }
    CRC32                       : Cardinal;
    CompressedSize              : Cardinal;
    UncompressedSize            : Cardinal;
    FileNameLength              : Word;
    ExtraFieldLength            : Word;
    FileName                    : String;
    ExtraField                  : String;
  end;

  TDataDescriptor = packed record
    CRC32            : Cardinal;
    CompressedSize   : Cardinal;
    UnCompressedSize : Cardinal;
  end;

  // struktura hlaviky v centrlnm adresi
  TFileHeader = packed record
    Version                     : Word;
    VersionNeedToExtract        : Word;
    Flag                        : Word;
    CompressionMethod           : Word;
    DateTime                    : Cardinal; { Time: Word; Date: Word; }
    CRC32                       : Cardinal;
    CompressedSize              : Cardinal;
    UncompressedSize            : Cardinal;
    FileNameLength              : Word;
    ExtraFieldLength            : Word;
    FileCommentLength           : Word;
    DiskNumberStart             : Word;
    InternalAttributes          : Word;
    ExternalAttributes          : Cardinal;
    Offset                      : Cardinal;
    FileName                    : String;
    ExtraField                  : String;
    FileComment                 : String;
    zSource                     : TStream;
    IsDirectory                 : Boolean;
  end;
  PFileHeader = ^TFileHeader;

  // Digitaln signatura
  TDigitalSignature = packed record
    Size                        : Word;
    SignatureData               : array of Byte;
  end;

  // ukonovac zznam centrlnho adrese
  TCentralDirectoryEnd = packed record
    NumberOfDisk                       : Word;
    NumberOfStartDisk                  : Word;
    TotalEntriesOnDisk                 : Word;
    TotalEntries                       : Word;   // celkov poet soubor
    Size                               : Cardinal;
    Offset                             : Cardinal;
    CommentLength                      : Word;
    Comment                            : string;
  end;
  PCentralDirectoryEnd = ^TCentralDirectoryEnd;

  THeaderList = class(TList)
  protected
    procedure Notify(Ptr: Pointer; Action: TListNotification); override;
  end;

  TZip = class
  private
    FHeaders: THeaderList;
    FSource:TStream;
    FSaveZip:TMemoryStream;
    FName:string;
    function GetFileCount:Cardinal;
    procedure SetSource(Value:TStream);
    procedure InitZIPFromStream(Stream:TStream);
  public
    constructor Create;
    destructor Destroy;override;
    procedure SaveZIPToStream(Dest:TStream);
    function FindHeader(const FileName:string):Integer;
    function FileSize(const FileName:string):Integer;
    // mazn souboru ze zipu
    procedure DeleteFile(const FileName:string);
    // Uloen souboru do ZIPu
    procedure SaveToZIP(const FileName:string;const Buffer;Count:Cardinal;Method:TCompressionLevel);
    // procedury pro proudov ten
    function OpenFile(const FileName:string):TStream;  // vrt stream, ze kterho je mon proudov st
    procedure CloseFile(Stream:TStream);        // uvoln stream
    // procedure pro naten celho souboru do pamti
    procedure LoadFile(const FileName:string;out Buff:pointer;out Bytes:Cardinal);
    property ZipSource:TStream read FSource write SetSource;
    property Headers:THeaderList read FHeaders;
    property FileCount:Cardinal read GetFileCount;
    property Name:string read FName write FName;
  end;

  TAlloc = function (AppData: Pointer; Items, Size: Integer): Pointer; register;
  TFree = procedure (AppData, Block: Pointer); register;
  // Internal structure.  Ignore.
  TZStreamRec = packed record
    next_in: PChar;       // next input byte
    avail_in: Integer;    // number of bytes available at next_in
    total_in: Integer;    // total nb of input bytes read so far

    next_out: PChar;      // next output byte should be put here
    avail_out: Integer;   // remaining free space at next_out
    total_out: Integer;   // total nb of bytes output so far

    msg: PChar;           // last error message, NULL if no error
    internal: Pointer;    // not visible by applications

    zalloc: TAlloc;       // used to allocate the internal state
    zfree: TFree;         // used to free the internal state
    AppData: Pointer;     // private data object passed to zalloc and zfree

    data_type: Integer;   //  best guess about the data type: ascii or binary
    adler: Integer;       // adler32 value of the uncompressed data
    reserved: Integer;    // reserved for future use
  end;

  // stream pro dekompresi ZIP souboru
  TZIPDecompressionStream = class(TStream)
  private
    FSource:TStream;
    FFirstPos:Integer;
    FPos:Integer;
    FZRec: TZStreamRec;
    FMethod:Integer;
    FBuffer: array [Word] of Char;
    FSize:Cardinal;
    procedure Reset;
  public
    constructor Create(Source: TStream;Method:Integer;Size:Cardinal);
    destructor Destroy; override;
    function Read(var Buffer; Count: Longint): Longint; override;
    function Write(const Buffer; Count: Longint): Longint; override;
    function Seek(Offset: Longint; Origin: Word): Longint; override;
  end;

  // stream pro stored metodu
  TStoredStream = class(TStream)
  private
    FSource:TStream;
    FSize:Integer;
    FFirstpos:Integer;
    FPos:Integer;
  public
    constructor Create(Source: TStream;Size:Integer);
    function Read(var Buffer; Count: Longint): Longint; override;
    function Write(const Buffer; Count: Longint): Longint; override;
    function Seek(Offset: Longint; Origin: Word): Longint; override;
  end;

  function Compare(Item1,Item2:Pointer):Integer;

implementation

const
  Z_NO_FLUSH      = 0;
  Z_PARTIAL_FLUSH = 1;
  Z_SYNC_FLUSH    = 2;
  Z_FULL_FLUSH    = 3;
  Z_FINISH        = 4;

  Z_OK            = 0;
  Z_STREAM_END    = 1;
  Z_NEED_DICT     = 2;
  Z_ERRNO         = (-1);
  Z_STREAM_ERROR  = (-2);
  Z_DATA_ERROR    = (-3);
  Z_MEM_ERROR     = (-4);
  Z_BUF_ERROR     = (-5);
  Z_VERSION_ERROR = (-6);

  Z_NO_COMPRESSION       =   0;
  Z_BEST_SPEED           =   1;
  Z_BEST_COMPRESSION     =   9;
  Z_DEFAULT_COMPRESSION  = (-1);

  Z_FILTERED            = 1;
  Z_HUFFMAN_ONLY        = 2;
  Z_DEFAULT_STRATEGY    = 0;

  Z_BINARY   = 0;
  Z_ASCII    = 1;
  Z_UNKNOWN  = 2;

  Z_DEFLATED = 8;

{$L obj\deflate.obj}
{$L obj\inflate.obj}
{$L obj\inftrees.obj}
{$L obj\trees.obj}
{$L obj\adler32.obj}
{$L obj\infblock.obj}
{$L obj\infcodes.obj}
{$L obj\infutil.obj}
{$L obj\inffast.obj}

procedure _tr_init; external;
procedure _tr_tally; external;
procedure _tr_flush_block; external;
procedure _tr_align; external;
procedure _tr_stored_block; external;
procedure inflate_blocks_new; external;
procedure inflate_blocks; external;
procedure inflate_blocks_reset; external;
procedure inflate_blocks_free; external;
procedure inflate_set_dictionary; external;
procedure inflate_trees_bits; external;
procedure inflate_trees_dynamic; external;
procedure inflate_trees_fixed; external;
procedure inflate_trees_free; external;
procedure inflate_codes_new; external;
procedure inflate_codes; external;
procedure inflate_codes_free; external;
procedure _inflate_mask; external;
procedure inflate_flush; external;
procedure inflate_fast; external;

procedure _memset(P: Pointer; B: Byte; count: Integer); cdecl;
begin
  FillChar(P^, count, B);
end;

procedure _memcpy(dest, source: Pointer; count: Integer); cdecl;
begin
  Move(source^, dest^, count);
end;

function Adler32(Adler:Cardinal;var Buffer;Length:Integer):Cardinal;external;

// deflate compresses data
function deflateInit_(var strm: TZStreamRec; level: Integer; version: PChar;
  recsize: Integer): Integer; external;
function deflate(var strm: TZStreamRec; flush: Integer): Integer; external;
function deflateEnd(var strm: TZStreamRec): Integer; external;

// inflate decompresses data
function inflateInit_(var strm: TZStreamRec; version: PChar;
  recsize: Integer): Integer; external;
function inflate(var strm: TZStreamRec; flush: Integer): Integer; external;
function inflateEnd(var strm: TZStreamRec): Integer; external;
function inflateReset(var strm: TZStreamRec): Integer; external;

function zlibAllocMem(AppData: Pointer; Items, Size: Integer): Pointer; register;
begin
//  GetMem(Result, Items*Size);
  Result := AllocMem(Items * Size);
end;

procedure zlibFreeMem(AppData, Block: Pointer); register;
begin
  FreeMem(Block);
end;

procedure CompressBuf(const InBuf: Pointer; InBytes: Integer;
                      out OutBuf: Pointer; out OutBytes: Integer;Level:Integer);
var
  strm: TZStreamRec;
  P: Pointer;
begin
  FillChar(strm, sizeof(strm), 0);
  strm.zalloc := zlibAllocMem;
  strm.zfree := zlibFreeMem;
  OutBytes := ((InBytes + (InBytes div 10) + 12) + 255) and not 255;
  GetMem(OutBuf, OutBytes);
  try
    strm.next_in := InBuf;
    strm.avail_in := InBytes;
    strm.next_out := OutBuf;
    strm.avail_out := OutBytes;
    deflateInit_(strm, Level, zlib_version, sizeof(strm));
    try
      while deflate(strm, Z_FINISH) <> Z_STREAM_END do
      begin
        P := OutBuf;
        Inc(OutBytes, 256);
        ReallocMem(OutBuf, OutBytes);
        strm.next_out := PChar(Integer(OutBuf) + (Integer(strm.next_out) - Integer(P)));
        strm.avail_out := 256;
      end;
    finally
      deflateEnd(strm);
    end;
    ReallocMem(OutBuf, strm.total_out);
    OutBytes := strm.total_out;
  except
    FreeMem(OutBuf);
    raise
  end;
end;

function Compare(Item1,Item2:Pointer):Integer;
begin
  Result:=Integer(Item1)-Integer(Item2);
  if (Item1=nil) or (Item2=nil) then exit;
  Result:=AnsiCompareStr(PFileHeader(Item1).FileName,
          PFileHeader(Item2).FileName);
end;

{ TZip }

constructor TZip.Create;
begin
  FName:='';
  FHeaders:=THeaderList.Create;
  FSaveZIP:=TMemoryStream.Create;
end;

destructor TZip.Destroy;
begin
  FHeaders.Free;
  FSaveZIP.Free;
  inherited;
end;

function TZip.GetFileCount: Cardinal;
begin
  Result:=FHeaders.Count;
end;

procedure TZip.InitZIPFromStream(Stream: TStream);
var name:array [0..MAX_PATH] of Char;
    sig:Cardinal;
    newheader:^TFileHeader;
    i,j:Cardinal;
    enddir:TCentralDirectoryEnd;
begin
  FHeaders.Clear;
  FSaveZIP.Size:=0;
  if (Stream = nil) or (Stream.Size<22) then exit;
  // naten hlaviek z centrlnho adrese, protoe
  // zde je vdy platn hodnota CRC32 a velikosti
  Stream.Seek(-22,soFromEnd);
  while true do begin
    // nalezen signatury - peskoen komente
    Stream.Read(sig,SizeOf(sig));
    if sig = ENDOFDIR_SIG then break;
    Stream.Seek(-SizeOf(sig)-1,soFromCurrent);
    if Stream.Position = 0 then exit;
  end;
  // naten zvrenho zznamu
  Stream.Read(enddir,SizeOf(enddir)-SizeOf(enddir.Comment));
  // nastaven na potek centrlnho adrese
  Stream.Seek(enddir.Offset,soFromBeginning);
  for i:=0 to enddir.TotalEntries -1 do begin
    // kontrola signatury
    Stream.Read(sig,SizeOf(sig));
    if sig<>HEADER_SIG then break;
    New(newheader);
    with newheader^ do begin
      Stream.Read(newheader^,HEADERSIZE-4);
      // peskome extrafield pole
      Stream.Seek(newheader^.ExtraFieldLength,soFromCurrent);
      // nateme jmno souboru
      Stream.Read(name,FileNameLength);
      FileName:=Copy(name,0,FileNameLength);
      // zmna '\' za '/', nebo v zipu je to tak uloeno
      for j:=1 to FileNameLength do
        if FileName[j] = '/' then
          FileName[j]:= '\';
      newheader.zSource:=FSource;
      IsDirectory:=FileName[FileNameLength]='\';
      // pidn do seznamu
      FHeaders.Add(newheader);
      // posun na dal zznam
      Stream.Seek(ExtraFieldLength+FileCommentLength,soFromCurrent);
    end;
  end;
  // setdn podle nzvu pro rychlej hledn
  FHeaders.Sort(@Compare);
end;

procedure TZip.SetSource(Value: TStream);
begin
  FSource:=Value;
  InitZIPFromStream(FSource);
end;

procedure TZip.SaveZIPToStream(Dest: TStream);
const b:Byte = 0;
var i,j:Integer;
    lh:TLocalHeader;
    fh:PFileHeader;
    offs,offsold:Cardinal;
    centrsize:Cardinal;
    enddir:TCentralDirectoryEnd;
begin
  // zpis zip do streamu
  // 1. zpis loklnch hlaviek a soubor
  if FileCount=0 then exit;
  for i:=0 to FileCount-1 do begin
    Dest.Write(LOCHEADER_SIG,SizeOf(LOCHEADER_SIG));
    with PFileHeader(FHeaders.Items[i])^ do begin
      lh.Version:=Version;
      lh.Flag:=Flag;
      lh.CompressionMethod:=CompressionMethod;
      lh.DateTime:=DateTime;
      lh.CRC32:=CRC32;
      lh.CompressedSize:=CompressedSize;
      lh.UncompressedSize:=UncompressedSize;
      lh.FileNameLength:=FileNameLength;
      lh.ExtraFieldLength:=0;
      Dest.Write(lh,LOCHEADERSIZE-4);
      Dest.Write(PChar(FileName)^,FileNameLength);
      // zapsn dat
      if CompressedSize = 0 then Continue;
      j:=Offset+LOCHEADERSIZE+FileNameLength;
      zSource.Seek(j,soFromBeginning);
      Dest.CopyFrom(zSource,CompressedSize);
    end;
  end;
  // 2.zapsn centrlnho adrese
  // te musme urit offset soubor, nebo zznamy v FHeaders mohou bt
  // v jinm poad, ne jak v pvodnm souboru

  // souasn urme velikost centrlnho adrese a jeho offset
  centrsize:=Dest.Position;
  offs:=0;
  for i:=0 to FileCount-1 do begin
    Dest.Write(HEADER_SIG,SizeOf(HEADER_SIG));
    fh:=PFileHeader(FHeaders.Items[i]);
    offsold:=fh^.Offset;
    fh^.Offset:=offs;
    Dest.Write(fh^,HEADERSIZE-4);
    Dest.Write(PChar(fh.FileName)^,fh.FileNameLength);
    fh^.Offset:=offsold;
    Inc(offs,LOCHEADERSIZE+fh^.CompressedSize+fh^.FileNameLength);
  end;
  with enddir do begin
    FillChar(enddir,SizeOf(enddir),0);
    Offset:=centrsize;
    centrsize:=Dest.Position-centrsize;
    Size:=centrsize;
    TotalEntriesOnDisk:=FileCount;
    TotalEntries:=FileCount;
  end;
  // 3.zpis ukonovacho zznamu
  Dest.Write(ENDOFDIR_SIG,SizeOf(ENDOFDIR_SIG));
  Dest.Write(enddir,DIRENDSIZE-4);
end;

function TZip.FindHeader(const FileName: string): Integer;
  function Interval(i1,i2:Integer):Integer;
  var centr:Integer;
      res:Integer;
  begin
    if (i1>i2) then begin
      Result:=-1;
      exit;
    end;
    centr:=(i1+i2) div 2;
    res:=AnsiCompareText(FileName,PFileHeader(FHeaders[centr])^.FileName);
    if res = 0 then
      Result:=centr
    else if i1=i2 then
      Result:=-1
    else if res<0 then
      Result:=Interval(i1,centr-1)
    else {res<0}
      Result:=Interval(centr+1,i2);
  end;
begin
  if FileName = '' then
    Result:=-1
  else
    Result:=Interval(0,FHeaders.Count-1);
end;

function TZip.FileSize(const FileName: string): Integer;
begin
  Result:=FindHeader(FileName);
  if Result<0 then exit;
  Result:=Integer(PFileHeader(FHeaders[Result])^.UncompressedSize);
end;

procedure TZip.CloseFile(Stream: TStream);
begin
  Stream.Free;
end;

procedure TZip.LoadFile(const FileName: string; out Buff: pointer;out Bytes:Cardinal);
var unzip:TStream;
begin
  Buff:=nil;
  Bytes:=0;
  unzip:=OpenFile(FileName);
  try
    Bytes:=FileSize(FileName);
    Buff:=AllocMem(Bytes);
    unzip.Read(Buff^,Bytes);
  finally
//    FreeMem(Buff);
    CloseFile(unzip);
  end;
end;

function TZip.OpenFile(const FileName: string): TStream;
var ind:Integer;
    fh:PFileHeader;
    method:Integer;
    src:TStream;
    extra:Word;
begin
  Result:=nil;
  ind:=FindHeader(FileName);
  if ind<0 then exit;
  fh:=PFileHeader(FHeaders[ind]);
  src:=fh.zSource;
  // POZOR! v zipu jsou dv hlaviky: lokln (ped kadm
  // souborem) a dal v centrlnm adresi. Ob mohou,
  // (nezvisle na sob) obsahovat zvltn pole, promnn dlky.
  // abychom nemuseli pi inicializaci skkat za lokln hlaviku
  // kvli zjitn dlky extrapole, zjiujeme tuto hodnotu te
  src.Seek(fh^.Offset+  LOCHEADERSIZE -2,soFromBeginning);
  // naten extrafieldlength
  src.Read(extra,SizeOf(extra));
  src.Seek(fh^.FileNameLength+extra,soFromCurrent);
  // jinou metodu ne deflate a bezkompresn neumme (bohuel)
  if fh^.CompressionMethod=8 then begin
    Case fh^.Flag and 6 of
      0:method:=Z_DEFAULT_COMPRESSION;
      6:method:=Z_BEST_SPEED;
    else
      method:=Z_BEST_COMPRESSION;
    end;
    Result:=TZIPDecompressionStream.Create(src,method,fh^.UncompressedSize);
  end else if fh^.CompressionMethod=0 then begin
    Result:=TStoredStream.Create(src,fh^.UncompressedSize);
  end;
end;

procedure TZip.SaveToZIP(const FileName: string; const Buffer;
  Count: Cardinal;Method:TCompressionLevel);
const
  Levels: array [TCompressionLevel] of ShortInt =
    (Z_NO_COMPRESSION, Z_BEST_SPEED, Z_DEFAULT_COMPRESSION, Z_BEST_COMPRESSION);
  Flags: array [TCompressionLevel] of ShortInt = (0,6,0,2);
var fh:PFileHeader;
    lh:TLocalHeader;
    csize:Integer;
    buf:PByte;
    i:Integer;
    temp:string;
begin
  buf:=nil;
  if FileName = '' then exit;
  // smazn , pokud soubor existuje
  DeleteFile(FileName);
  // inicializace FileHeader struktury
  New(fh);
  FillChar(fh^,SizeOf(fh^),0);
  fh.FileName:=FileName;
  FSaveZIP.Seek(0,soFromEnd);
  with fh^ do begin
    FileNameLength:=Length(FileName);
    Version:=20;
    VersionNeedToExtract:=20;
    DateTime:=DateTimeToFileDate(Time);
    UncompressedSize:=Count;
    zSource:=FSaveZIP;
    CRC32:=CalcCRC.Crc32(0,Buffer,Count);
    Flag:=Flags[method];
    Offset:=FSaveZIP.Position;
    IsDirectory:=FileName[FileNameLength]='\';
  end;
  FSaveZIP.Write(lh,LOCHEADERSIZE);
  // zmna '/' za '\', nebo v zipu je to tak uloeno
  temp:=Copy(FileName,0,fh.FileNameLength);
  for i:=1 to fh.FileNameLength do
    if temp[i] = '\' then
      temp[i] := '/';
  FSaveZIP.Write(PChar(temp)^,fh.FileNameLength);
  if Method=clNone then begin
    fh.CompressedSize:=Count;
    FSaveZIP.Write(Buffer,Count);
  end else begin
    fh.CompressionMethod:=8;
    try
      CompressBuf(@Buffer,Count,pointer(buf),csize,Levels[Method]);
      fh.CompressedSize:=csize-6; // bez hlaviky [2] a kontrl. soutu [4]
      Inc(buf,2);
      FSaveZIP.Write(buf^,fh.CompressedSize);
      Dec(buf,2);
    finally
      Freemem(buf);
    end;
  end;
  // lokln hlaviku nen teba nastavovat, protoe ji nepouvme
  FHeaders.Add(fh);
  FHeaders.Sort(@Compare);
end;

procedure TZip.DeleteFile(const FileName: string);
var ind:Integer;
    locsize:Integer;
    i:Integer;
    src:pointer;
    dest:pointer;
begin
  // mae se pouze z FHeaders
  ind:=FindHeader(FileName);
  if ind = -1 then exit;
  with PFileHeader(FHeaders[ind])^ do
    if zSource = FSaveZIP then begin
      // smaeme i ze streamu
      // 1. upravme offsety ostatnch soubor
      locsize:=LOCHEADERSIZE+CompressedSize+FileNameLength;  // ExtraFiledLength nen teba, protoe pi ukldn do FSaveZIP je vdy nulov
      for i:=0 to FileCount-1 do
        with PFileHeader(FHeaders[i])^ do
          if (zSource=FSaveZIP) and (Offset>PFileHeader(FHeaders[ind])^.Offset) then
            Dec(Offset,locsize);
      // 2.pesuneme zznamy nachzejc se za prv mazanm
      dest:=Pointer(Cardinal(FSaveZIP.Memory)+Offset);
      src:=Pointer(Integer(dest)+locsize);
      Move(src^,dest^,FSaveZIP.Size-Offset-locsize);
      // 3.zmenme potebnou pamt
      FSaveZIP.Size:=FSaveZIP.Size-locsize;
    end;
  FHeaders.Delete(ind);
end;

{ THeaderList }

procedure THeaderList.Notify(Ptr: Pointer; Action: TListNotification);
begin
  Case Action of
    lnDeleted:
      if Assigned(Ptr) then
        Dispose(Ptr);
  end;
end;

{ TZIPDecompressionStream }

constructor TZIPDecompressionStream.Create;
begin
  inherited Create;
  FMethod:=Method;
  FSource:=Source;
  FFirstPos:=Source.Position;
  FPos:=FFirstPos;
  FSize:=Size;
  Reset;
end;

destructor TZIPDecompressionStream.Destroy;
begin
  inflateEnd(FZRec);
  inherited Destroy;
end;

function TZIPDecompressionStream.Read(var Buffer; Count: Integer): Longint;
begin
  FZRec.next_out := @Buffer;
  FZRec.avail_out := Count;

  while (FZRec.avail_out > 0) do begin
    if FZRec.avail_in = 0 then begin
      FZRec.avail_in := FSource.Read(FBuffer, sizeof(FBuffer));
      if FZRec.avail_in = 0 then begin
        FillChar(FBuffer,SizeOf(FBuffer),0);
        FZRec.avail_in:=SizeOf(FBuffer);
      end;
      FZRec.next_in := FBuffer;
      FPos := FSource.Position;
    end;
    // pokud dojde na konec streamu, zkou nast Adler32 - vdy vyvol chybu, proto ignorujeme
    if inflate(FZRec, 0)<0 then break;
  end;
  Result := Count;
end;

procedure TZIPDecompressionStream.Reset;
var sig:Word;
begin
  // nastaven ukazatele na zatek + naten signatury
  with FZRec do begin
    // pokud jsme nco naetli musme jet navc resetovat
    if FFirstPos<>FPos then begin
      inflateReset(FZRec);
      FPos:=FFirstPos;
      FSource.Position:=FPos;
    end;
    zalloc := zlibAllocMem;
    zfree := zlibFreeMem;
    avail_in := 0;
    next_out := nil;
    avail_out := 0;
    inflateInit_(FZRec, zlib_version, sizeof(FZRec));
    if FMethod = Z_BEST_COMPRESSION then
      sig:=$DA78
    else
      sig:=$0178;
    next_in := @sig;
    avail_in := 2;
    // naten chybjc signatury
    inflate(FZRec, 0);
    next_in := FBuffer;
  end;
end;

function TZIPDecompressionStream.Seek(Offset: Integer;
  Origin: Word): Longint;
var
  I: Integer;
  Buf: array [0..4095] of Char;
begin
  Case Origin of
    soFromCurrent:Inc(Offset,FPos-FFirstpos);
    soFromEnd:Inc(Offset,FSize);
  end;
  if Offset = 0 then begin
    Reset;
  end else if (Offset - FZRec.total_out) > 0 then begin
    Dec(Offset, FZRec.total_out);
    if Offset > 0 then begin
      for I := 1 to Offset div sizeof(Buf) do
        ReadBuffer(Buf, sizeof(Buf));
      ReadBuffer(Buf, Offset mod sizeof(Buf));
    end;
  end else begin
    // posun zpt vyeme jako posun na zatek plus posun dopedu
    Seek(0,soFromBeginning);
    Seek(Offset,soFromBeginning);
  end;
  Result := FZRec.total_out;
end;

function TZIPDecompressionStream.Write(const Buffer;
  Count: Integer): Longint;
begin
  raise EStreamError.Create('Invalid operation');
end;

{ TStoredStream }

constructor TStoredStream.Create(Source: TStream; Size: Integer);
begin
  inherited Create;
  FSource:=Source;
  FSize:=Size;
  FFirstpos:=FSource.Position;
  FPos:=FFirstpos;
end;

function TStoredStream.Read(var Buffer; Count: Integer): Longint;
begin
  if FSource.Position<>FPos then
    FSource.Position:=FPos;
  if (Count+FPos)>(FSize+FFirstPos) then
    Count:=FSize+FFirstPos-FPos;
  FSource.Read(Buffer,Count);
  FPos:=FSource.Position;
  Result:=Count;
end;

function TStoredStream.Seek(Offset: Integer; Origin: Word): Longint;
begin
  Case Origin of
    soFromBeginning:
        FSource.Seek(FFirstPos+Offset,soFromBeginning);
    soFromCurrent:
        FSource.Seek(Offset,soFromCurrent);
    soFromEnd:
        FSource.Seek(FFirstPos+Offset+FSize,soFromBeginning);
  end;
  Result:=FSource.Position-FFirstPos;
  if Result>FSize then
    Result:=FSize;
  if Result<0 then
    Result:=0;
  FPos:=Result+FFirstPos;
end;

function TStoredStream.Write(const Buffer; Count: Integer): Longint;
begin
  raise EStreamError.Create('Invalid operation');
end;

end.



